@@ -45,7 +45,7 @@ class Status(enum.Enum):
             return "missed"
         if self == Status.BOTH:
             return "both"
-        raise UserWarning("Found unknown line coverage status: {}".format(self))
+        raise UserWarning(f"Found unknown line coverage status: {self}")
 
     def __str__(self):
         """A string representation of line coverage"""
@@ -66,8 +66,7 @@ class Status(enum.Enum):
             return Status.MISSED
         if status.lower() == "both":
             return Status.BOTH
-        raise UserWarning("Found unknown line coverage status: {}"
-                          .format(status))
+        raise UserWarning(f"Found unknown line coverage status: {status}")
 
     def combine(self, status):
         """Combine line coverage"""
@@ -226,7 +225,7 @@ def merge_coverage_data(coverage_list):
     try:
         coverage and RAW_COVERAGE_DATA(coverage)
     except voluptuous.Error as error:
-        raise UserWarning("Error merging coverage data: {}".format(error)) from error
+        raise UserWarning(f"Error merging coverage data: {error}") from error
     return coverage
 
 
@@ -307,17 +306,14 @@ def load_json(json_file):
     """Load json file produced by make-coverage"""
 
     json_data = parse.parse_json_file(json_file)
-    if json_data is None:
-        logging.info("Found empty json coverage data.")
-        return None
+    if not json_data:
+        raise UserWarning(f"Failed to load json coverage data: {json_file}")
 
     coverage = repair_json(json_data[JSON_TAG]['coverage'])
     try:
         RAW_COVERAGE_DATA(coverage)
     except voluptuous.Error as error:
-        raise UserWarning(
-            "Error loading json coverage data: {}: {}".format(json_file, error)
-        ) from error
+        raise UserWarning(f"Invalid json coverage data: {json_file}: {error}") from error
     return coverage
 
 def repair_json(coverage):
@@ -357,31 +353,28 @@ def load_cbmc_json(json_file, root):
 
     json_data = parse.parse_json_file(json_file)
     if not json_data:
-        logging.info("Expected coverage data in json file %s, found none",
-                     json_file)
-        return None
+        raise UserWarning(f"Failed to load json coverage data: {json_file}")
 
-    goal_list = [entry for entry in json_data if 'goals' in entry]
-    if len(goal_list) != 1:
-        logging.info("Expected 1 block of goal data in json file %s, found %s",
-                     json_file, len(goal_list))
-        return None
-    goals = goal_list[0]
+    try:
+        goals_list = [entry for entry in json_data if 'goals' in entry]
+        assert len(goals_list) == 1
+        goals = goals_list[0]['goals']
+    except AssertionError as error:
+        raise UserWarning(
+            f"Failed to locate coverage goals in json coverage data: {json_file}") from error
 
     coverage = {}
-    for goal in goals["goals"]:
-        description = goal["description"]
+    for goal in goals:
+        lines = (parse_basicBlockLines(goal.get("basicBlockLines")) or
+                 parse_description(goal.get("description")))
         status = goal["status"]
-        location = goal["sourceLocation"]
-        wkdir = srcloct.json_srcloc_wkdir(location)
-        coverage = add_coverage_data(coverage, description, status, wkdir, root)
+        wkdir = srcloct.json_srcloc_wkdir(goal["sourceLocation"])
+        coverage = add_coverage_data(coverage, lines, status, wkdir, root)
 
     try:
         RAW_COVERAGE_DATA(coverage)
     except voluptuous.Error as error:
-        raise UserWarning(
-            "Error loading cbmc json coverage data: {}: {}".format(json_file, error)
-        ) from error
+        raise UserWarning(f"Invalid json coverage data: {json_file}: {error}") from error
     return coverage
 
 ################################################################
@@ -402,50 +395,40 @@ def load_cbmc_xml(xml_file, root):
 
     xml = parse.parse_xml_file(xml_file)
     if not xml or xml is None:   # Why is 'xml is None' required?
-        logging.info("Expected coverage data in xml file %s, found none",
-                     xml_file)
-        return None
+        raise UserWarning(f"Failed to load xml coverage data: {xml_file}")
 
     coverage = {}
     for goal in xml.iter("goal"):
-        description = goal.get("description")
+        lines = (parse_basic_block_lines(goal.find("basic_block_lines")) or
+                 parse_description(goal.get("description")))
         status = goal.get("status")
-        location = goal.find("location")
-        wkdir = srcloct.xml_srcloc_wkdir(location)
-        coverage = add_coverage_data(coverage, description, status, wkdir, root)
+        wkdir = srcloct.xml_srcloc_wkdir(goal.find("location"))
+        coverage = add_coverage_data(coverage, lines, status, wkdir, root)
 
     try:
         RAW_COVERAGE_DATA(coverage)
     except voluptuous.Error as error:
-        raise UserWarning(
-            "Error loading cbmc xml coverage data: {}: {}".format(xml_file, error)
-        ) from error
+        raise UserWarning(f"Invalid xml coverage data: {xml_file}: {error}") from error
     return coverage
 
 ################################################################
 # Parse coverage data
 
-def add_coverage_data(coverage, description, status, wkdir, root):
+def add_coverage_data(coverage, lines, status, wkdir, root):
     """Add to coverage the coverage data reported for a coverage goal"""
 
-    if not description:
-        logging.debug("Found a missing coverage description in coverage data.")
-        return coverage
-
     # Warning: What follows assumes that all relative paths appearing
-    # in a basic block's coverage description are relative to the same
-    # working directory (the working directory in the source location
-    # labeling the basic block's coverage goal that contains the
-    # coverage description).  This is probably true.  A project can
+    # in a coverage goal's list of source lines are relative to the
+    # same working directory (the working directory in the coverage
+    # goal's source location).  This is probably true.  A project can
     # avoid this issue altogether, however, by invoking goto-cc on
     # absolute paths, ensuring that coverage descriptions contain only
     # absolute paths and no relative paths.
 
     hit = parse_coverage_status(status)
-    locations = parse_coverage_description(description)
-    locations = relative_locations(locations, wkdir, root)
+    lines = relative_locations(lines, wkdir, root)
 
-    for path, func, line in locations:
+    for path, func, line in lines:
         coverage = update_coverage(coverage, path, func, line, hit)
     return coverage
 
@@ -487,59 +470,104 @@ def update_coverage(coverage, path, func, line, status):
     return coverage
 
 ################################################################
-# Parse the basic block of code in a coverage goal
+# Extract the source lines contributing to a coverage goal's basic block
+#
+# Each coverage goal is a basic block in a goto program. Each
+# instruction in a basic block has a source location identifying the
+# source line that generated the instruction.
+#
+# CBMC encodes these source lines in the text of the goal description.
+# Later versions of CBMC also encode these source lines a xml or json
+# subelements of the goal in CBMC's xml or json coverage output.
 
-def parse_coverage_description(string):
-    """Parse a coverage description and extract the basic block covered."""
+def parse_lines(string):
+    """Extract line numbers from the string encoding of line numbers used in coverage output"""
 
-    # The coverage description has the form "block N (lines BASIC_BLOCK)"
-    match = re.match(r'block [0-9]+ \(lines (.*)\)', string)
-    if not match:
-        logging.info("Found unparsable basic block description: %s", string)
-        return None
+    # string is an encoding of a set of line numbers like
+    #   "1,3,6" -> {1, 3, 6}
+    #   "1-3,6,20-32" -> {1, 2, 3, 6, 20, 21, 22}
 
-    return parse_basic_block(match.group(1))
-
-def parse_basic_block(basic_block):
-    """Parse a basic block of coverage"""
-
-    # The basic block is a sequence of chunks CHUNK1;CHUNK2;CHUNK3
-    return [location
-            for chunk in basic_block.split(';')
-            for location in parse_chunk(chunk)]
-
-def parse_chunk(chunk):
-    """Parse a chunk of coverage"""
-
-    # cbmc has added whitespace to coverage descriptions
-    chunk = chunk.strip()
-
-    # The chunk has the form FILE:FUNCTION:LINES
-    # Don't use split in order to allow FUNCTION to contain ':'
-    first_colon, last_colon = chunk.find(':'), chunk.rfind(':')
-    filename = chunk[:first_colon]
-    function = chunk[first_colon + 1:last_colon]
-    lines = chunk[last_colon + 1:]
-    return [(filename, function, line) for line in parse_lines(lines)]
-
-def parse_lines(lines):
-    """Parse the lines in a chunk of coverage"""
-
-    # groups of lines are separated by commas.
-    # a group of lines is either a single line N or a range of lines N-M
-    line_list = []
-    for line_range in lines.split(','):
-        bounds = line_range.split('-')
+    lines = set()
+    for group in string.split(','):
+        bounds = group.split('-')
         if len(bounds) == 1:
-            # Add the single line N
-            line_list.append(int(bounds[0]))
+            lines.add(int(bounds[0]))
+        elif len(bounds) == 2:
+            lines.update(range(int(bounds[0]), int(bounds[1])+1))
         else:
-            # Add the range of lines N-M = N, N+1, ..., M
-            line_list.extend(range(int(bounds[0]), int(bounds[1])+1))
+            raise UserWarning(f"Unexpected encoding of line numbers: {string}")
+    return sorted(lines)
 
-    # don't worry about duplicates in line_list: if a line is included
-    # twice it will be set to the same coverage status twice
-    return line_list
+def parse_description(description):
+    """Extract basic block source lines from a coverage goal's textual description"""
+
+    try:
+        # description is "block N (lines BASIC_BLOCK)"
+        basic_block = re.match(r'block [0-9]+ \(lines (.*)\)', description).group(1)
+
+        # basic_block is
+        #   chunk1;chunk2;chunk3
+        # each chunk is
+        #   test.c:foo:3,6-10
+        # each function name foo from rust may include embedded semicolons like
+        #   main::foo
+        lines = [(fyle, func, line)
+                 for chunk in basic_block.split(';')
+                 for fyle, func_lines in [chunk.split(':', 1)]  # fyle:func:lines -> fyle,func:lines
+                 for func, lines in [func_lines.rsplit(':', 1)] # func:lines -> func,lines
+                 for line in parse_lines(lines)]
+        assert all(fyle and func and line for fyle, func, line in lines)
+        return lines
+    except (AttributeError, ValueError, AssertionError) as error:
+        # AttributeError after match(): 'NoneType' object has no attribute 'group'
+        # ValueError after after split()/rsplit(): not enough values to unpack
+        # ValueError in parse_lines(): invalid literal for int()
+        raise UserWarning(f'Unexpected coverage goal description: "{description}"') from error
+
+def parse_basic_block_lines(basic_block_lines):
+    """Extract basic block source lines from a coverage goal's xml subelement"""
+
+    if basic_block_lines is None:
+        return []
+
+    try:
+        # basic_block_lines is xml
+        #   <basic_block_lines>
+        #     <line file="test.c" function="foo">3,6-10</line>
+        #     ...
+        #   </basic_block_lines>
+        lines = [(fyle, func, line)
+                 for bbl in basic_block_lines.iter("line")
+                 for fyle, func, lines in [(bbl.get("file"), bbl.get("function"), bbl.text)]
+                 for line in parse_lines(lines)]
+        assert all(fyle and func and line for fyle, func, line in lines)
+        return lines
+    except (ValueError, AssertionError) as error:
+        # ValueError in parse_lines(): invalid literal for int()
+        raise UserWarning(f'Unexpected coverage goal xml data: "{basic_block_lines}"') from error
+
+def parse_basicBlockLines(basicBlockLines): # pylint: disable=invalid-name
+    """Extract basic block source lines from a coverage goal's json data"""
+
+    if basicBlockLines is None:
+        return []
+
+    try:
+        # basicBlockLines is json
+        #   "basicBlockLines": {
+        #     "test.c": {
+        #       "main": "16,17"
+        #     }
+        #   }
+        lines = [(fyle, func, line)
+                 for fyle, fyle_data in basicBlockLines.items()
+                 for func, lines in fyle_data.items()
+                 for line in parse_lines(lines)]
+        assert all(fyle and func and line for fyle, func, line in lines)
+        return lines
+    except (ValueError, AssertionError) as error:
+        # ValueError in parse_lines(): invalid literal for int()
+        raise UserWarning(f'Unexpected coverage goal json data: "{basicBlockLines}"') from error
 
 ################################################################
 # Parse the hit/miss status in a coverage goal
@@ -568,15 +596,14 @@ def make_coverage(args):
     if viewer_coverage:
         if filet.all_json_files(viewer_coverage):
             return CoverageFromJson(viewer_coverage)
-        fail("Expected json files: {}".format(viewer_coverage))
+        fail(f"Expected json files: {viewer_coverage}")
 
     if cbmc_coverage and srcdir:
         if filet.all_json_files(cbmc_coverage):
             return CoverageFromCbmcJson(cbmc_coverage, srcdir)
         if filet.all_xml_files(cbmc_coverage):
             return CoverageFromCbmcXml(cbmc_coverage, srcdir)
-        fail("Expected json files or xml files, not both: {}"
-             .format(cbmc_coverage))
+        fail(f"Expected json files or xml files, not both: {cbmc_coverage}")
 
     logging.info("make-coverage: nothing to do: need "
                  "cbmc coverage checking results and --srcdir or "
