@@ -4,8 +4,8 @@
 """Version number."""
 
 NAME = "CBMC viewer"
-NUMBER = "3.0.1"
-VERSION = "{} {}".format(NAME, NUMBER)
+NUMBER = "3.4"
+VERSION = f"{NAME} {NUMBER}"
 
 def version(display=False):
     """The version of cbmc viewer."""
